"""

This script generates subfigures a,b,c,d,e from Figure 2 of the manuscript
'Learning models of quantum systems from experiments'.
Note Figure 2f is a representation of a single QMLA instance which requires
the complete QMLA code base to reproduce, so is not included here.

The results come from 8 runs, each of >100 QMLA instances, with different intentions.
* 1 run for experimental data using |+>|+'> probe.
* 1 run for simulated data, assuming the true model using |+>|+'> probe.
* 1 run for simulated data, assuming the true model using random probes.
* 5 runs for simulated data, each assuming a differnt true model, using random probes.

The results of these runs are stored in combined_results.csv.
run_info.csv contains information about the different runs,
such as the true model and whether experimental data was used.

The analysis here imports those CSV files into pandas dataframes.
We then preprocess the dataframes to have only the data pertinent to each plot.

Plots are generated by standalone functions.
These functions are then called within subplots to generate figure_2_analysis.pdf.

"""

import numpy as np
import pandas as pd
import scipy as sp
import math as mth
from IPython.display import display, Math, Latex
from functools import partial
import datetime
import os as os
import time as time
import importlib as imp
import logging as logging
import warnings as warnings
import sys

import tabulate
import pickle
import networkx as nx
import qinfer as qi
import qutip as qt

import matplotlib.pyplot as plt
import matplotlib.cbook as cb
from matplotlib.colors import colorConverter, Colormap
from matplotlib.patches import FancyArrowPatch, Circle, ArrowStyle
import matplotlib.colors as colors
from matplotlib.colors import ListedColormap
import matplotlib.patches as mpatches
from matplotlib.path import Path
import matplotlib.cm as cm
import matplotlib.colors as mpc
from matplotlib.gridspec import GridSpec
from matplotlib import collections
from matplotlib import ticker
from matplotlib import transforms
from matplotlib.ticker import FuncFormatter, MaxNLocator
from matplotlib.ticker import Formatter
from matplotlib import rc
from matplotlib import rcParams
from matplotlib.gridspec import GridSpec
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

###########################################
# Miscellaneous functions, borrowed from QMLA
# code base, used while processing data
###########################################
def get_constituent_names_from_name(name):
    verbose_naming_mechanism_terms = ['T', 'P', 'M']

    if np.any(
        [t in name for t in verbose_naming_mechanism_terms]
    ):
        return verbose_naming_mechanism_separate_terms(name)
    else:
        return name.split('+')


def verbose_naming_mechanism_separate_terms(name):
    t_str, p_str, max_t, max_p = get_t_p_strings(name)
    if(max_t >= max_p):
        # if more T's than P's in name,
        # it has only one constituent.
        return [name]
    else:
        # More P's indicates a sum at the highest dimension.
        return name.split(p_str)


def get_t_p_strings(name):
    """
    Find largest instance of consecutive P's and T's.
    Return those instances and lengths of those instances.
    """
    t_str = ''
    p_str = ''
    while name.count(t_str + 'T') > 0:
        t_str = t_str + 'T'

    while name.count(p_str + 'P') > 0:
        p_str = p_str + 'P'

    max_t = len(t_str)
    max_p = len(p_str)

    return t_str, p_str, max_t, max_p


###########################################
# Functions which act on dataframes to give plots
###########################################
# (a) Note this is just an imshow of a PNG image.

# (b) Fittings rates with inset R squared histogram.
def fitting_rates(
    results_df,
    runs_to_evaluate,
    run_info,
    ax
):
    """
    Acts on the entire results dataframe to produce a plot showing
    the fitting rate as a function of parameter cardinality difference
    between the true and champion model.

    """
    subset_results = results_df[
        (results_df.RunIdx.isin(runs_to_evaluate))
    ]

    percentage_rate = 100 / len(subset_results)
    run_info['ConstituentTermsTrue'] = run_info.TrueModel.apply(
        get_constituent_names_from_name
    )
    run_info['NumParamsTrue'] = run_info.ConstituentTermsTrue.apply(len)

    subset_results['ConstituentTermsChamp'] = subset_results.NameAlphabetical.apply(
        get_constituent_names_from_name
    )
    subset_results['NumParamsChamp'] = subset_results.ConstituentTermsChamp.apply(
        len)

    subset_results = subset_results.join(
        run_info, on='RunIdx', rsuffix="NumParamsTrue")
    subset_results.FittingMismatch = subset_results.NumParamsTrue - \
        subset_results.NumParamsChamp

    mismatches = dict(subset_results.FittingMismatch.value_counts())
    difference_from_true = sorted(mismatches.keys())

    underfit = [m for m in difference_from_true if m < 0]
    misfit = [m for m in difference_from_true if m == 0]
    overfit = [m for m in difference_from_true if m > 0]
    correct = [0]

    underfit_scores = [mismatches[d] * percentage_rate for d in underfit]
    misfit_scores = [mismatches[d] * percentage_rate for d in misfit]
    overfit_scores = [mismatches[d] * percentage_rate for d in overfit]
    correct_scores = [int(subset_results.CorrectModel.sum()) * percentage_rate]

    ax.barh(
        underfit,
        underfit_scores,
        label='Under',
        color=fitcolors[2]
    )

    ax.barh(
        misfit,
        misfit_scores,
        label='Mis',
        color=fitcolors[0]
    )
    ax.barh(
        correct,
        correct_scores,
        label='Correct',
        color=fitcolors[3]
    )

    ax.barh(
        overfit,
        overfit_scores,
        label='Over',
        color=fitcolors[1]
    )

    ax.set_ylabel(
        r'$N_p - N_p^{\prime}$',
        **axis_font
    )
    ax.set_xlabel(
        'Win rate (%)',
        **axis_font
    )

    ax.set_yticks([-4, -3, -2, -1, 0, 1])
    ax.set_ylim(-6, 3)
    plt.gca().invert_yaxis()
    ax.legend(
        fontsize=legend_fontsize,
        loc=4,  # lower right,
        ncol=2
    )
    ax.tick_params(
        which='both',
        labelsize=legend_fontsize,
        pad=axis_label_padding,
    )

    return subset_results


def r_squared_histogram(
    simulation_df,
    ax
):
    """

    Acts on sub-dataframe of only simulated data.
    Find the R-squared value of the champion models
    and produces a histogram.
    This is displayed as an inset of the fitting rates plot.

    """

    r_squared_values = np.array(list(simulation_df['FinalRSquared']))
    r_median = np.median(r_squared_values)
    print("R squared median:", r_median)
    font = {'family': 'normal',
            'weight': 'normal',
            'size': 11}

    rc('font', **font)

    reduced_r_sq = [r for r in r_squared_values if r > 0]
    num_less_than_zero = len(r_squared_values) - len(reduced_r_sq)
    negatives = [-0.1] * num_less_than_zero
    reduced_r_sq.extend(negatives)
    bins = list(np.arange(-0.2, 1.1, 0.1))

    bin_fills = ax.hist(reduced_r_sq, bins=bins)

    ax.set_title(
        '$R^2$ Histogram',
        fontdict=inset_caption_font,
        y=0.6
    )
    ax.set_xlim(-0.1, 1)

    max_y = max(bin_fills[0])
    ax.set_ylim(0, max_y * 1.2)
    ax.tick_params(
        axis='y',
        direction='out',
        labelleft=True,
        labelright=False,
        length=20,
    )
    xt = [-0.05, 0.05, 0.5, 0.95]
    xt_labels = ['<0', 0, 0.5, 1]
    ax.set_xticks(xt)
    ax.set_xticklabels(xt_labels)
    ax.set_yticks([50, 100, 150])
    vline_lw = 8
    ax.tick_params(
        which='both',
        labelsize=legend_fontsize * .75,
        pad=axis_label_padding * .75
    )
    ax.axvline(0, lw=vline_lw)
    ax.axvline(r_median, lw=vline_lw, ls='--', c='red')

# (c) Win rates: number of QMLA instances won by each model
def model_win_rates(
    dataframes,
    top_models,
    ax
):
    """
    Given a list of dataframes, corresponding to the individual runs desired to show,
    this finds the number of QMLA instances (out of 100) which each model won,
    and presents the scores for models within the top_models list given.
    """
    combined_df = pd.concat(
        dataframes
    )

    runs = list(combined_df.RunIdx.unique())
    scores = {}
    for r in runs:
        scores[r] = dict(
            combined_df[
                (combined_df['RunIdx'] == r)
            ].ChampLatex.value_counts()
        )

    # set up widths
    width = 0.35  # the width of the bars
    ind = 1.6 * np.arange(len(top_models))  # the x locations for the groups

    widths = {}
    b = 0
    w = 0
    for r in runs:
        idx = runs.index(run)
        b += 1
        if b % 2 == 0:
            w += width
        a = np.power(-1, b)
        widths[r] = a * w

    for r in runs:
        wins = [
            scores[r][m] if m in list(
                scores[r].keys()) else 0 for m in top_models]

        ax.barh(
            ind - widths[r],
            wins,
            width,
            linewidth=0,
            label=run_labels[r],
            color=run_colour_dict[r]
        )

    ax.set_xlabel(
        'Win rate (%)',
        **axis_font
    )
    ax.set_yticks(ind)
    ax.set_yticklabels(
        top_models,
        minor=False,
        **axis_font
    )
    ax.tick_params(
        axis='both',
        labelsize=legend_fontsize,
        pad=axis_label_padding
    )

    ax.legend(
        fontsize=legend_fontsize
    )
    return scores


# (d) Volume of chosen models
def volume_plot(
    df,
    ax,
    top_models=None,
    num_top_models=4,
    return_volume_df=False,
):
    """
    Given a list top_models (or, failing that, the num_top_models to extract
    from the dataframe), this plot shows the volume of the posterior distribution
    for the chosen models at each epoch of their learning procedure.

    """
    if top_models is None:
        top_models = experimental_df.ChampLatex.value_counts().index.tolist()[
            :num_top_models]

    volumes_df = pd.DataFrame()
    qids = [int(i) for i in sorted(df.index.tolist())]
    for k in qids:
        row = df.loc[int(k)]
        a = [
            pd.Series(
                {
                    'RunIdx': row.RunIdx,
                    'ChampLatex': row.ChampLatex,
                    'Instance': int(row.QID),
                    'Epoch': e,
                    'Volume': v
                }
            )
            for e, v in zip(eval(row.Epochs)[::10], eval(row.RawVolumes)[::10])
        ]

        volumes_df = volumes_df.append(
            pd.DataFrame(a),
            ignore_index=True
        )

    volume_groups = volumes_df.groupby(
        by=[
            'ChampLatex',
            'Epoch'
        ]
    )['Volume']
    medians = volume_groups.median()
    upper = volume_groups.quantile(0.835)
    lower = volume_groups.quantile(0.165)
    epochs = sorted(volumes_df.Epoch.unique())
    instances = sorted(volumes_df.Instance.unique())
    for mod in top_models:

        multi_indices = [(mod, e) for e in epochs]
        vol_medians = [medians[mi] for mi in multi_indices]
        vol_upper = [upper[mi] for mi in multi_indices]
        vol_lower = [lower[mi] for mi in multi_indices]
        ax.semilogy(
            epochs,
            vol_medians,
            label=mod,
            color=model_colour_dict[mod],
            lw=4
        )
        ax.set_ylabel('Volume', **axis_font)
        ax.set_xlabel('Epoch', **axis_font)
        ax.fill_between(
            epochs,
            vol_lower,
            vol_upper,
            color=model_colour_dict[mod],
            alpha=0.2
        )
    ax.legend(
        fontsize=legend_fontsize,
        ncol=2, loc=1,
        #         ncol=1, loc=3
    )

    ax.tick_params(
        which='both',
        labelsize=legend_fontsize,
        pad=axis_label_padding)

    if return_volume_df == True:
        return volumes_df, multi_indices

# (e) Dynamics of experimental data with the reproduced dynamics of the champion model
def dynamics_plot(
    dynamics_plot_run,
    dynamics_plot_model,
    expectation_values,
    true_expectation_values,
    ax,
    lower_quantile=0.165,
    upper_quantile=0.835
):
    """

    Given expectation values dataframe, and true_expectation_values (experimental),
    averages the expectation values of the champion models for the run on experimental data.
    Only the dynamics_plot_model is plotted; in this case we choose the model which wins
    most frequently.
    Shaded area shows the lower_quantile and upper_quantile provided to the fnc,
    to show the variance in the predictive power.
    """
    run_id = dynamics_plot_run

    grouped_expec_vals = expectation_values.groupby(
        by=['RunIdx', 'ChampLatex', 'time']
    )['ExpVal']

    expec_vals_median = grouped_expec_vals.median()
    expec_vals_upper_quantile = grouped_expec_vals.quantile(upper_quantile)
    expec_vals_lower_quantile = grouped_expec_vals.quantile(lower_quantile)

    true_vals = true_expectation_values.groupby(
        by=['RunIdx', 't']
    ).mean()
    times = sorted(
        true_expectation_values[true_expectation_values.RunIdx ==
                                dynamics_plot_run].t.unique()
    )
    # get true data
    true_multi_indices = [(run_id, t) for t in times]
    true = [true_vals.loc[tmi].TrueExpVal for tmi in true_multi_indices]

    # get learned data
    multi_indices = [(run_id, dynamics_plot_model, t) for t in times]
    medians = [expec_vals_median.loc[mi] for mi in multi_indices]
    upper = [expec_vals_upper_quantile.loc[mi] for mi in multi_indices]
    lower = [expec_vals_lower_quantile.loc[mi] for mi in multi_indices]

    ax.scatter(
        times,
        true,
        label="Experimental Data".format(run_id),
        s=primary_font_size * 5,
        color=true_model_colour
    )

    ax.plot(
        times,
        medians,
        label=dynamics_plot_model,
        lw=primary_font_size * 0.35,
        color=model_colour_dict[dynamics_plot_model]
    )
    ax.fill_between(
        times,
        lower,
        upper,
        color=model_colour_dict[dynamics_plot_model],
        alpha=0.3,

    )

    ax.legend(
        fontsize=legend_fontsize
        #         **legend_font
    )
    ax.set_xlabel(
        r'Time ($\mu s$)',
        **axis_font
    )
    ax.set_yticks([0, 0.5, 1])
    ax.tick_params(
        which='both',
        labelsize=legend_fontsize,
        pad=axis_label_padding)
    ax.set_ylabel(
        'Expec. Val',
        **axis_font
    )






###########################################
# Import datasets from CSV files.
###########################################

shared_results_path = os.getcwd()
combined_results_path = os.path.abspath(
    os.path.join(
        shared_results_path,
        'combined_results.csv'))
run_info_path = os.path.abspath(
    os.path.join(
        shared_results_path,
        'run_info.csv'))
exp_val_path = os.path.abspath(
    os.path.join(
        shared_results_path,
        'expectation_values.csv'))
true_exp_val_path = os.path.abspath(
    os.path.join(
        shared_results_path,
        'true_expectation_values.csv'))
volumes_path = os.path.abspath(
    os.path.join(
        shared_results_path,
        'volumes.csv'))

# read CSV into dataframes
results_df = pd.read_csv(combined_results_path)
run_info = pd.read_csv(run_info_path)
true_expectation_values = pd.read_csv(true_exp_val_path)
run_info.set_index('RunIdx', inplace=True)

# Notes run index which unifies results_df and run_info:
# 0 : experimental data
# 2 : simulated experiment
# others are for simulations of random systems.

# Since some runs have >100 instances,
# remove extra results at random and keep exactly 100 instances per run

run_instance_cap = 100
runs = [int(r) for r in list(results_df.RunIdx.unique())]

extraneous_results = pd.Series(
    results_df.RunIdx.value_counts() - run_instance_cap
)
removed_indices = []
for run_id in extraneous_results.index.tolist():
    num_extra = extraneous_results[run_id]
    random_sample = results_df[
        (results_df.RunIdx == run_id)
    ].sample(num_extra).index.tolist()
    removed_indices.extend(random_sample)

    results_df.drop(random_sample, inplace=True)

# Generate separate dataframes for experimental and simulation plots
num_top_models_to_plot = 4
overall_top_models = results_df.ChampLatex.value_counts().index.tolist()[
    :num_top_models_to_plot]

experimental_df = results_df[
    (results_df['RunIdx'] == 0)
][0:100]  # so that exactly 100 runs considered

top_models = experimental_df.ChampLatex.value_counts().index.tolist()[
    :num_top_models_to_plot]
top_models = sorted(top_models, key=len)
top_model_df = experimental_df[
    (experimental_df['ChampLatex'].isin(top_models))
]


experiment_simulation_df = results_df[
    (results_df['RunIdx'] == 2)
][0:100]  # so that exactly 100 runs considered

# runs to compare
run_labels = {
    0: 'Experiment',
    2: 'Simulation'
}

# Isolate expectation values for dynamics plot only
run_for_dynamics = 0
dynamics_model = '$\\hat{S}_{x,y,z}\\hat{A}_{z}$'
desired_results = results_df[
    (results_df['RunIdx'].isin([run_for_dynamics]))
    &
    (results_df['ChampLatex'].isin([dynamics_model]))
]
ind = desired_results.index.tolist()
expectation_values = pd.DataFrame()
for k in ind:
    row = results_df.iloc[k]
    a = [
        pd.Series(
            {
                'RunIdx': row.RunIdx,
                'ChampLatex': row.ChampLatex,
                'time': t,
                'ExpVal': e
            }
        )
        for t, e in zip(eval(row.ExpValTimes), eval(row.RawExpVals))
    ]
    expectation_values = expectation_values.append(
        pd.DataFrame(a),
        ignore_index=True
    )


# Generate colour dict for consistency across plots
# Note run_colour_dict, model_colour_dict are used in every function for
# consistency.
color_scheme = cm.viridis  # can be changed to user's preference.

min_color = 0.3
max_color = 0.8
cm_subsection = np.linspace(min_color, max_color, 4)
fitcolors = [color_scheme(x) for x in cm_subsection]
true_model_colour = 'red'
cm_subsection = np.linspace(min_color, max_color, len(top_models))
colours = [color_scheme(x) for x in cm_subsection]

model_colour_dict = {}
for mod in top_models:
    model_colour_dict[mod] = colours[top_models.index(mod)]

compared_runs = list(run_labels.keys())
cm_subsection = np.linspace(min_color, max_color, len(compared_runs))
colours = [color_scheme(x) for x in cm_subsection]

run_colour_dict = {}
for run in compared_runs:
    run_colour_dict[run] = colours[compared_runs.index(run)]


###########################################
# Configuration used by all plots for consistency
# Note all controls are defined relative to primary_font_size,
# so changing that alone can change the output plot.
###########################################

primary_font_size = 10
fig_base_size = primary_font_size * 1.5
fig_size = (fig_base_size * 1.6, fig_base_size * 1.1)


rc('font', family='Microsoft Sans Serif')
rc('xtick', labelsize=primary_font_size)
rc('ytick', labelsize=primary_font_size)

rc('legend', fontsize=primary_font_size - 1)


title_font = {
    'fontname': 'Microsoft Sans Serif',
    'size': str(
        2 * primary_font_size),
    'color': 'black',
    'weight': 'normal'}
axis_font = {
    'fontname': 'Microsoft Sans Serif',
    'size': str(
        2 * primary_font_size),
    'color': 'black',
    'weight': 'normal'}
ticks_font = {
    'fontname': 'Microsoft Sans Serif',
    'size': str(
        2 * primary_font_size)}
legend_font = {
    'fontname': 'Microsoft Sans Serif',
    'fontsize': str(
        2 * primary_font_size)}
caption_font = {
    'fontname': 'Microsoft Sans Serif',
    'fontsize': str(
        1.8 * primary_font_size),
    'weight': 'bold'}
inset_caption_font = {
    'fontname': 'Microsoft Sans Serif',
    'fontsize': str(
        1.8 * primary_font_size),
    'weight': 'normal'}
legend_fontsize = 1.5 * primary_font_size
axis_label_padding = primary_font_size / 2

rcParams['xtick.direction'] = 'in'
rcParams['ytick.direction'] = 'in'

###########################################
# Call plotting functions on axes within subplot to generate combined image.
###########################################


fig = plt.figure(
    figsize=fig_size,
    tight_layout=True
)


spacing = 0.35
title_padding = 20

gs = GridSpec(
    nrows=4,
    ncols=3,
    width_ratios=[1, 1, 1.3, ],
    height_ratios=[1, 1, 1, 0.3]
)

# Image for fig (a)
path_to_bloch_sphere = os.path.abspath(
    os.path.join(
        shared_results_path,
        'bloch_sphere.png'

    )
)
bloch = plt.imread(path_to_bloch_sphere)
ax = fig.add_subplot(gs[0, 0])
ax.imshow(bloch)
ax.axis('off')
ax.set_title(
    'a',
    loc='left',
    fontdict=caption_font,
    pad=title_padding
)


# Fitness and r squared histogram
ax = fig.add_subplot(gs[0, 1])
simulation_df = fitting_rates(
    results_df,
    run_info=run_info,
    runs_to_evaluate=[4, 5, 6, 7, 8],
    ax=ax
)
ax.set_title(
    'b',
    loc='left',
    fontdict=caption_font,
    pad=title_padding
)
inset_axis = inset_axes(ax, width=primary_font_size/2.5, height=primary_font_size/7, loc=1)
r_squared_histogram(
    simulation_df,
    inset_axis
)


# Success Rates
ax = fig.add_subplot(gs[1, 0])
scores_as_in_plot = model_win_rates(
    dataframes=[experimental_df, experiment_simulation_df],
    top_models=top_models,
    ax=ax
)
ax.set_title(
    'c',
    loc='left',
    fontdict=caption_font,
    pad=title_padding
)


# Volumes
ax = fig.add_subplot(gs[1, 1])
volume_plot(
    df=top_model_df,
    ax=ax,
    num_top_models=4,
    return_volume_df=False
)
ax.set_title(
    'd',
    loc='left',
    fontdict=caption_font,
    pad=title_padding
)


# Dynamics
ax = fig.add_subplot(
    gs[2, 0:2],
)
dynamics_plot(
    dynamics_plot_run=0,
    dynamics_plot_model='$\\hat{S}_{x,y,z}\\hat{A}_{z}$',
    expectation_values=expectation_values,
    true_expectation_values=true_expectation_values,
    ax=ax,
    lower_quantile=0.165,
    upper_quantile=0.835
)
ax.set_title(
    'e',
    loc='left',
    fontdict=caption_font,
    pad=title_padding
)

# Single QMLA instance tree
# Note not available here
# ax = fig.add_subplot(
#     gs[0:4, 2],
# )

# single_qmd_directory_name = os.path.abspath(
#     os.path.join(
#         shared_results_path, 'single_instance_classes/'
#     )
# )
# qmd_path = os.path.abspath(
#     os.path.join(
#         single_qmd_directory_name,
#         'qmd_class_002.p'
#     )
# )

# single_qmd_tree_plot(
#     ax = ax,
#     single_qmd_directory_name = single_qmd_directory_name,
#     qmd_path = qmd_path,
#     bayes_factor_cap = 50,
#     label_fontsize = legend_fontsize,
#     edge_colormap = plt.cm.YlGnBu,
# )
# # ax.axvline(0)
# # ax.axhline(0)
# ax.set_title(
#     'f',
#     loc='left',
#     fontdict = caption_font,
#     pad = title_padding
# )

fig.tight_layout()
fig.savefig('figure_2_analysis.pdf', bbox_inches='tight')


